Require Import floyd.proofauto.
Import ListNotations.
Local Open Scope logic.

Require Import sha.spec_sha.
Require Import sha_lemmas.
Require Import sha.HMAC_functional_prog.
Require Import sha.HMAC256_functional_prog.
Require Import hmac_common_lemmas.

Require Import sha.hmac.
Require Import spec_hmac.

Module Type HMAC_ABSTRACT_SPEC.

(*"Ordinary" abstract states contains a key and some data*)
Inductive HABS := hABS: forall (key data:list Z), HABS.

(*The mpred REP (hABS k d) c expresses that pointer value c holds
  a ctx representing the situation where we have hmac-ed
  data a with key k, and are prepeared to accept additional data.*)
Parameter REP: HABS -> val -> mpred.

(*We have two additional protocol states. FULL k c holds after we return
  from hmac-final (so we can't hmac-updeate more data into ctx).
  It's the precondition of calls to hmac_init with argument key==null, ie 
  the case where we want to reuse the key k in the next round of hmac. *)
Parameter FULL: list Z -> val -> mpred.

(*EMPTY c captures the situation where we either haven't provided any key yet,
  or want to use an old ctx, but reinitialize its key. It occurs explicitly
  in the precondition of init_call with argument key==Vptr b i*)
Parameter EMPTY: val -> mpred.

(*We can turn a memory block of hmac_ctx size into an EMPTY abstract HMAC REP*)
Parameter mkEmpty: forall v, field_compatible t_struct_hmac_ctx_st [] v -> 
memory_block Tsh (sizeof t_struct_hmac_ctx_st) v |-- EMPTY v.

(* We can prematurely terminate sequences of hmac-update by simply declaring
   an updateable ctx FULL*)
Parameter REP_FULL: forall key data c, REP (hABS key data) c |-- FULL key c.

(*We can also "wipe" a ctx, ie forget/erase any key material from ctx*)
Parameter FULL_EMPTY: forall key c, FULL key c |-- EMPTY c.

Parameter EMPTY_isptr: forall c, EMPTY c |-- !!isptr c.

Lemma FULL_isptr: forall key c, FULL key c |-- !!isptr c.
Proof.
  intros.
  eapply derives_trans.
  apply FULL_EMPTY.
  apply EMPTY_isptr.
Qed.

Lemma REP_isptr: forall key data c, REP (hABS key data) c |-- !!isptr c.
Proof.
  intros.
  eapply derives_trans.
  apply REP_FULL.
  apply FULL_isptr.
Qed.

(****** Next, here are the new specs of openssl's API functions *************)
(* I have split the two cases for hmac_init. mbedtls separates these situations in its API:
   mbedtls_md_hmac_reset prepares to authenticate a new message with the same key;
   mbedtls_md_hmac_starts sets the key and prepares to authenticate a new message. *)

Definition hmac_reset_spec :=
  DECLARE _HMAC_Init (*Naphat: you'll probably have DECLARE mbedtls_hmac_reset here, and the
                       body of your wrapper function is a call to hmac_init with key==null.*)
   WITH c : val, l:Z, key:list Z, kv:val 
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _key OF tptr tuchar,
         _len OF tint ]
         PROP ()
         LOCAL (temp _ctx c; temp _key nullval; temp _len (Vint (Int.repr l));
                gvar sha._K256 kv)
         SEP (FULL key c; K_vector kv)
  POST [ tvoid ] 
     PROP ()
     LOCAL ()
     SEP (REP (hABS key nil) c; K_vector kv).

Definition hmac_starts_spec :=
  DECLARE _HMAC_Init (*Naphat: you'll probably have DECLARE mbedtls_hmac_starts here, and the
                       body of your wrapper function is a call to hmac_init with the nonnull key*)
   WITH c : val, l:Z, key:list Z, kv:val, b:block, i:Int.int
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _key OF tptr tuchar,
         _len OF tint ]
         PROP (has_lengthK l key)
         LOCAL (temp _ctx c; temp _key (Vptr b i); temp _len (Vint (Int.repr l));
                gvar sha._K256 kv)
         SEP (EMPTY c; data_block Tsh key (Vptr b i); K_vector kv)
  POST [ tvoid ] 
     PROP ()
     LOCAL ()
     SEP (REP (hABS key nil) c; data_block Tsh key (Vptr b i); K_vector kv).

Definition hmac_update_spec :=
  DECLARE _HMAC_Update
   WITH key: list Z, c : val, d:val, data:list Z, data1:list Z, kv:val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st, 
         _data OF tptr tvoid, 
         _len OF tuint]
         PROP (0 <= Zlength data1 <= Int.max_unsigned /\
               Zlength data1 + Zlength data + 64 < two_power_pos 61) 
         LOCAL (temp _ctx c; temp _data d; temp  _len (Vint (Int.repr (Zlength data1)));
                gvar sha._K256 kv)
         SEP(REP (hABS key data) c; data_block Tsh data1 d; K_vector kv)
  POST [ tvoid ] 
          PROP () 
          LOCAL ()
          SEP(REP (hABS key (data++data1)) c; 
              data_block Tsh data1 d; K_vector kv).

Definition hmac_final_spec :=
  DECLARE _HMAC_Final
   WITH data:list Z, key:list Z, c : val, md:val, shmd: share, kv:val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _md OF tptr tuchar ]
       PROP (writable_share shmd) 
       LOCAL (temp _md md; temp _ctx c;
              gvar sha._K256 kv)
       SEP(REP (hABS key data) c; K_vector kv;
           memory_block shmd 32 md)
  POST [ tvoid ] 
          PROP () 
          LOCAL ()
          SEP(K_vector kv; FULL key c;
              data_block shmd (HMAC256 data key) md).

(*Maybe this is not needed in mbedtls?*)
Definition hmac_cleanup_spec :=
  DECLARE _HMAC_cleanup
   WITH key: list Z, c : val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st ]
         PROP () 
         LOCAL (temp _ctx c)
         SEP(FULL key c)
  POST [ tvoid ]  
          PROP () 
          LOCAL ()
          SEP(EMPTY c).

(* Finally, we have prrofs that the openssl implementations satisfy these specs*)

Parameter body_hmac_final: semax_body HmacVarSpecs HmacFunSpecs 
                           f_HMAC_Final hmac_final_spec. 

Parameter body_hmac_update: semax_body HmacVarSpecs HmacFunSpecs 
                            f_HMAC_Update hmac_update_spec. 

Parameter body_hmac_starts: semax_body HmacVarSpecs HmacFunSpecs 
                             f_HMAC_Init hmac_starts_spec.
 
Parameter body_hmac_reset: semax_body HmacVarSpecs HmacFunSpecs 
                              f_HMAC_Init hmac_reset_spec. 

Parameter body_hmac_cleanup: semax_body HmacVarSpecs HmacFunSpecs 
                             f_HMAC_cleanup hmac_cleanup_spec.

End HMAC_ABSTRACT_SPEC.

Lemma haslengthK_simple: forall l, 0 < l <= Int.max_signed -> l * 8 < two_p 64.
intros. 
assert (l < Int.half_modulus). unfold Int.max_signed in H. omega. clear H.
rewrite Int.half_modulus_power in H0. 
assert (Int.zwordsize = 32) by reflexivity. rewrite H in *; clear H. simpl in *.
rewrite two_power_pos_equiv in *. 
assert (l * 8 < 2^31 * 8) by omega. clear H0.
eapply Z.lt_trans. eassumption. clear H. cbv; trivial.
Qed.

Module OPENSSL_HMAC_ABSTRACT_SPEC <: HMAC_ABSTRACT_SPEC.
Inductive HABS := hABS: forall (key data:list Z), HABS.

Definition abs_relate (a: HABS) (r: hmacstate) : Prop :=
  match a with hABS key data => 
    hmac_relate (hmacUpdate data (hmacInit key)) r
  end. 

Definition REP (a: HABS) (c: val) : mpred :=
   EX r:hmacstate, 
    (!!(abs_relate a r) && data_at Tsh t_struct_hmac_ctx_st r c).

Definition FULL key c:mpred :=
   (*!!(has_lengthK (Zlength key) key) &&*) EX h:_, hmacstate_PreInitNull key h c.

Definition EMPTY c : mpred := data_at_ Tsh t_struct_hmac_ctx_st c.

Lemma mkEmpty v: field_compatible t_struct_hmac_ctx_st [] v -> 
  memory_block Tsh (sizeof t_struct_hmac_ctx_st) v |-- EMPTY v.
Proof. intros. unfold EMPTY. rewrite data_at__memory_block. entailer!. Qed.

Lemma REP_FULL key data c: REP (hABS key data) c |-- FULL key c.
Proof. unfold REP, FULL. Intros r.
  unfold hmacstate_PreInitNull. simpl in H.
  destruct H as [mREL [iREL [oREL [iLEN oLEN]]]].
  Exists (hmacUpdate data (hmacInit key)) r (fst r).
  apply andp_right.
    apply prop_right. simpl. intuition.
  apply derives_refl'. f_equal. destruct r as [md [IS OS]]. simpl. reflexivity.
Qed.

Lemma FULL_EMPTY key c: FULL key c |-- EMPTY c.
Proof. unfold FULL, EMPTY.
 unfold hmacstate_PreInitNull. Intros h r v. 
 apply data_at_data_at_.
Qed.

Lemma EMPTY_isptr c: EMPTY c |-- !!isptr c. 
Proof. unfold EMPTY. entailer!. Qed.

Lemma FULL_isptr key c: FULL key c |-- !!isptr c.
Proof.
  eapply derives_trans.
  apply FULL_EMPTY.
  apply EMPTY_isptr.
Qed.

Lemma REP_isptr key data c: REP (hABS key data) c |-- !!isptr c.
Proof.
  eapply derives_trans.
  apply REP_FULL.
  apply FULL_isptr.
Qed.

(************************ Abstract specifications of HMAC_init *******************************************************)

Definition hmac_reset_spec :=
  DECLARE _HMAC_Init
   WITH c : val, l:Z, key:list Z, kv:val (*, d:list Z*)
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _key OF tptr tuchar,
         _len OF tint ]
         PROP ()
         LOCAL (temp _ctx c; temp _key nullval; temp _len (Vint (Int.repr l));
                gvar sha._K256 kv)
         SEP (FULL key c; K_vector kv)
  POST [ tvoid ] 
     PROP ()
     LOCAL ()
     SEP (REP (hABS key nil) c; K_vector kv).

Definition hmac_starts_spec :=
  DECLARE _HMAC_Init
   WITH c : val, l:Z, key:list Z, kv:val, b:block, i:Int.int
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _key OF tptr tuchar,
         _len OF tint ]
         PROP (has_lengthK l key)
         LOCAL (temp _ctx c; temp _key (Vptr b i); temp _len (Vint (Int.repr l));
                gvar sha._K256 kv)
         SEP (EMPTY c; data_block Tsh key (Vptr b i); K_vector kv)
  POST [ tvoid ] 
     PROP ()
     LOCAL ()
     SEP (REP (hABS key nil) c; data_block Tsh key (Vptr b i); K_vector kv).

Definition hmac_update_spec :=
  DECLARE _HMAC_Update
   WITH key: list Z, c : val, d:val, data:list Z, data1:list Z, kv:val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st, 
         _data OF tptr tvoid, 
         _len OF tuint]
         PROP (0 <= Zlength data1 <= Int.max_unsigned /\
               Zlength data1 + Zlength data + 64 < two_power_pos 61) 
         LOCAL (temp _ctx c; temp _data d; temp  _len (Vint (Int.repr (Zlength data1)));
                gvar sha._K256 kv)
         SEP(REP (hABS key data) c; data_block Tsh data1 d; K_vector kv)
  POST [ tvoid ] 
          PROP () 
          LOCAL ()
          SEP(REP (hABS key (data++data1)) c; 
              data_block Tsh data1 d; K_vector kv).

Definition hmac_final_spec :=
  DECLARE _HMAC_Final
   WITH data:list Z, key:list Z, c : val, md:val, shmd: share, kv:val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st,
         _md OF tptr tuchar ]
       PROP (writable_share shmd) 
       LOCAL (temp _md md; temp _ctx c;
              gvar sha._K256 kv)
       SEP(REP (hABS key data) c; K_vector kv;
           memory_block shmd 32 md)
  POST [ tvoid ] 
          PROP () 
          LOCAL ()
          SEP(K_vector kv;
              FULL key c;
              data_block shmd (HMAC256 data key) md).

Definition hmac_cleanup_spec :=
  DECLARE _HMAC_cleanup
   WITH key: list Z, c : val
   PRE [ _ctx OF tptr t_struct_hmac_ctx_st ]
         PROP () 
         LOCAL (temp _ctx c)
         SEP(FULL key c)
  POST [ tvoid ]  
          PROP () 
          LOCAL ()
          SEP(EMPTY c).

Require Import verif_hmac_final.
Require Import verif_hmac_update.
Require Import verif_hmac_init.
Require Import verif_hmac_cleanup.

Lemma body_hmac_reset: semax_body HmacVarSpecs HmacFunSpecs 
       f_HMAC_Init hmac_reset_spec. 
Proof.
start_function.
rename lvar0 into pad. rename lvar1 into ctxkey.
abbreviate_semax.
apply semax_pre with (P':=EX h1:hmacabs, 
  (PROP  ()
   LOCAL  (lvar _ctx_key (tarray tuchar 64) ctxkey;
   lvar _pad (tarray tuchar 64) pad; temp _ctx c; temp _key nullval;
   temp _len (Vint (Int.repr l));  gvar sha._K256 kv)
   SEP  (data_at_ Tsh (tarray tuchar 64) ctxkey;
   data_at_ Tsh (tarray tuchar 64) pad; K_vector kv;
   initPre c nullval h1 l key))). 
{ unfold FULL. Intros h1. Exists h1. (*red in H.*)  entailer!. }
Intros h1.
eapply semax_post.
2: apply (initbodyproof Espec c nullval l key kv h1 pad ctxkey).
  intros. apply andp_left2. apply frame_ret_assert_derives.
  apply function_body_ret_assert_derives.
  old_go_lower.
  entailer!.
  unfold hmacstate_, REP. Intros r. Exists r. entailer!.
  red. rewrite hmacUpdate_nil. assumption. 
Qed.

Lemma body_hmac_final: semax_body HmacVarSpecs HmacFunSpecs 
       f_HMAC_Final hmac_final_spec. 
Proof.
start_function.
rename lvar0 into buf.
unfold REP, abs_relate. Intros r.
destruct H as [mREL [iREL [oREL [iLEN oLEN]]]].
eapply semax_pre_post.
  3: apply (finalbodyproof Espec c md shmd kv buf (hmacUpdate data (hmacInit key)) SH).
  
  apply andp_left2. unfold hmacstate_. Exists r. old_go_lower. entailer!.

  intros. apply andp_left2. apply frame_ret_assert_derives.
  apply function_body_ret_assert_derives.
  rewrite <- hmac_sound. unfold FULL.
  change (hmacFinal (hmacUpdate data (hmacInit key))) with (hmac key data).
  Exists (fst (hmac key data)). old_go_lower. entailer!.
  eapply hmacstate_PostFinal_PreInitNull; reflexivity.
Qed.

Lemma body_hmac_update: semax_body HmacVarSpecs HmacFunSpecs 
       f_HMAC_Update hmac_update_spec. 
Proof.
start_function.
destruct H as [Prop1 Prop2].
eapply semax_pre_post.
  3: apply (updatebodyproof Espec c d (Zlength data1) data1 kv (hmacUpdate data (hmacInit key))).

  apply andp_left2. old_go_lower. entailer!.
  unfold REP, hmacstate_. Intros r. Exists r. entailer!.

  intros. apply andp_left2. apply frame_ret_assert_derives.
  apply function_body_ret_assert_derives. 
  rewrite hmacUpdate_app. old_go_lower. entailer!.
  unfold REP, hmacstate_.
  Intros r. Exists r. entailer!.

  split; trivial. split; trivial. simpl.
  unfold innerShaInit, s256a_len.
  rewrite Zlength_app, Zlength_mkArgZ, map_length, mkKey_length, Min.min_idempotent.
  simpl. rewrite (Z.add_comm 64), <- Z.mul_add_distr_r, Z.add_assoc. 
  assert (Tpp: (two_power_pos 64 = two_power_pos 61 * 8)%Z) by reflexivity.
  rewrite Tpp.  
  apply Zmult_lt_compat_r. omega. trivial. 
Qed.  

Lemma body_hmac_starts: semax_body HmacVarSpecs HmacFunSpecs 
       f_HMAC_Init hmac_starts_spec. 
Proof.
start_function.
rename lvar0 into pad. rename lvar1 into ctxkey.
unfold EMPTY. 
remember (HMACabs (S256abs nil nil) (S256abs nil nil) (S256abs nil nil)) as hdummy.
eapply semax_pre_post.
Focus 3. apply (initbodyproof Espec c (Vptr b i) l key kv hdummy pad ctxkey).
  apply andp_left2. old_go_lower. entailer!.
  intros. apply andp_left2. apply frame_ret_assert_derives.
  apply function_body_ret_assert_derives. old_go_lower. entailer!. 
   unfold hmacstate_, REP. Intros r. Exists r. entailer!.
   red. rewrite hmacUpdate_nil. assumption.
Qed.

Lemma body_hmac_cleanup: semax_body HmacVarSpecs HmacFunSpecs 
       f_HMAC_cleanup hmac_cleanup_spec.
Proof.
start_function.
unfold FULL. Intros h.
assert_PROP (field_compatible t_struct_hmac_ctx_st [] c).
{ unfold hmacstate_PreInitNull. Intros r v. entailer!. }
eapply semax_pre_post.
  3: apply (cleanupbodyproof1 Espec c h).
  Exists key. apply andp_left2. apply derives_refl. 

  intros. apply andp_left2.  apply frame_ret_assert_derives.
  apply function_body_ret_assert_derives. old_go_lower. entailer!.
  unfold EMPTY. 
  rewrite <- memory_block_data_at_. simpl. unfold data_block.
  clear. simpl. apply andp_left2. apply data_at_memory_block. trivial.
Qed. 

End OPENSSL_HMAC_ABSTRACT_SPEC.
